"""
Helper functions to create and delete queues on SQS.
"""


def create_standard_queue(
    env,
    queue_name,
    message_retention_period=None,
    visibility_timeout=None,
    redrive_policy=None,
):
    """
    Create a new standard queue
    """
    attrs = {}
    kwargs = {"QueueName": env.get_sqs_queue_name(queue_name), "Attributes": attrs}
    if message_retention_period is not None:
        attrs["MessageRetentionPeriod"] = str(message_retention_period)
    if visibility_timeout is not None:
        attrs["VisibilityTimeout"] = str(visibility_timeout)
    if redrive_policy is not None:
        attrs["RedrivePolicy"] = redrive_policy.__json__()
    ret = env.sqs_client.create_queue(**kwargs)
    return ret["QueueUrl"]


def create_fifo_queue(
    env,
    queue_name,
    content_based_deduplication=False,
    message_retention_period=None,
    visibility_timeout=None,
    redrive_policy=None,
):
    """
    Create a new FIFO queue. Note that queue name has to end with ".fifo"

    - "content_based_deduplication" turns on automatic content-based
      deduplication of messages in the queue

    - redrive_policy can be None or an object, generated with
      redrive_policy() method of SQS. In the latter case if defines the
      way failed messages are processed.
    """
    attrs = {"FifoQueue": "true"}
    kwargs = {"QueueName": env.get_sqs_queue_name(queue_name), "Attributes": attrs}
    if content_based_deduplication:
        attrs["ContentBasedDeduplication"] = "true"
    if message_retention_period is not None:
        attrs["MessageRetentionPeriod"] = str(message_retention_period)
    if visibility_timeout is not None:
        attrs["VisibilityTimeout"] = str(visibility_timeout)
    if redrive_policy is not None:
        attrs["RedrivePolicy"] = redrive_policy.__json__()
    ret = env.sqs_client.create_queue(**kwargs)
    return ret["QueueUrl"]


def delete_queue(env, queue_name):
    """
    Delete the queue
    """
    env.sqs_resource.get_queue_by_name(
        QueueName=env.get_sqs_queue_name(queue_name)
    ).delete()



from contextlib import contextmanager
from typing import TYPE_CHECKING, Callable

from sqs_workers.codecs import DEFAULT_CONTENT_TYPE
from sqs_workers.utils import bind_arguments

if TYPE_CHECKING:
    from sqs_workers.queue import JobQueue


class AsyncTask(object):
    def __init__(self, queue: "JobQueue", job_name: str, processor: Callable):
        self.queue = queue
        self.job_name = job_name
        self.processor = processor
        self.__doc__ = processor.__doc__

    def __call__(self, *args, **kwargs):
        raise RuntimeError(
            "Async task {0.queue.name}.{0.job_name} called synchronously (probably, "
            "by mistake). Use either AsyncTask.run(...) to run the task synchronously "
            "or AsyncTask.delay(...) to add it to the queue".format(self)
        )

    def __repr__(self):
        return "<%s %s.%s>" % (self.__class__.__name__, self.queue.name, self.job_name)

    def run(self, *args, **kwargs):
        """
        Run the task synchronously.
        """
        if self.queue.batching_policy.batching_enabled:
            if len(args) > 0:
                raise TypeError("Must use keyword arguments only for batch read queues")
            kwargs = bind_arguments(self.processor, [[kwargs]], {})
            return self.processor(**kwargs)
        else:
            kwargs = bind_arguments(self.processor, args, kwargs)
            return self.processor(**kwargs)

    @contextmanager
    def batch(self):
        """
        Context manager to add jobs in batch.
        """
        with self.queue.add_batch():
            yield

    def delay(self, *args, **kwargs):
        """
        Run the task asynchronously.
        """
        _content_type = kwargs.pop("_content_type", DEFAULT_CONTENT_TYPE)
        _delay_seconds = kwargs.pop("_delay_seconds", None)
        _deduplication_id = kwargs.pop("_deduplication_id", None)
        _group_id = kwargs.pop("_group_id", None)

        if self.queue.batching_policy.batching_enabled:
            if len(args) > 0:
                raise TypeError("Must use keyword arguments only for batch read queues")
        else:
            kwargs = bind_arguments(self.processor, args, kwargs)

        return self.queue.add_job(
            self.job_name,
            _content_type=_content_type,
            _delay_seconds=_delay_seconds,
            _deduplication_id=_deduplication_id,
            _group_id=_group_id,
            **kwargs
        )

    def bake(self, *args, **kwargs):
        """
        Create a baked version of the async task, which contains the reference
        to a queue and task, as well as all arguments which needs to be passed
        to it.
        """
        return BakedAsyncTask(self, args, kwargs)


class BakedAsyncTask(object):
    def __init__(self, async_task, args, kwargs):
        self.async_task = async_task
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        self.async_task(*self.args, **self.kwargs)

    def delay(self):
        self.async_task.delay(*self.args, **self.kwargs)

    def __repr__(self):
        return "BakedAsyncTask(%r, ...)" % self.async_task



from sqs_workers.backoff_policies import (
    DEFAULT_BACKOFF,
    IMMEDIATE_RETURN,
    ConstantBackoff,
    ExponentialBackoff,
)
from sqs_workers.exceptions import SQSError
from sqs_workers.memory_sqs import MemorySession
from sqs_workers.queue import JobQueue, RawQueue
from sqs_workers.sqs_env import SQSEnv
from sqs_workers.sqs_manage import (
    create_fifo_queue,
    create_standard_queue,
    delete_queue,
)

__all__ = [
    "DEFAULT_BACKOFF",
    "IMMEDIATE_RETURN",
    "ConstantBackoff",
    "ExponentialBackoff",
    "SQSError",
    "MemorySession",
    "JobQueue",
    "RawQueue",
    "SQSEnv",
    "create_fifo_queue",
    "create_standard_queue",
    "delete_queue",
]



"""
In memory mockup implementation of essential parts of boto3 SQS objects.

Used for faster and more predictable processing in tests.

Lacks some features of "real sqs", and some other features implemented
ineffectively.

- Redrive policy doesn't work
- There is no differences between standard and FIFO queues
- FIFO queues don't support content-based deduplication
- Delayed tasks executed ineffectively: the task is gotten from the queue,
  and if the time hasn't come yet, the task is put back.
- API can return slightly different results

Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html
"""
import datetime
import uuid
from typing import Any, Dict, List

import attr


@attr.s
class MemoryAWS(object):
    """
    In-memory AWS as a whole.
    """

    client: "MemoryClient" = attr.ib(
        repr=False,
        default=attr.Factory(lambda self: MemoryClient(self), takes_self=True),
    )
    resource: "ServiceResource" = attr.ib(
        repr=False,
        default=attr.Factory(lambda self: ServiceResource(self), takes_self=True),
    )
    queues: List["MemoryQueue"] = attr.ib(factory=list)

    def create_queue(self, QueueName, Attributes):
        queue = MemoryQueue(self, QueueName, Attributes)
        self.queues.append(queue)
        return queue

    def delete_queue(self, QueueUrl):
        self.queues = [queue for queue in self.queues if queue.url != QueueUrl]


@attr.s
class MemorySession(object):
    """
    In memory AWS session.
    """

    aws = attr.ib(repr=False, factory=MemoryAWS)

    def client(self, service_name):
        assert service_name == "sqs"
        return self.aws.client

    def resource(self, service_name):
        assert service_name == "sqs"
        return self.aws.resource


@attr.s
class MemoryClient(object):

    aws = attr.ib(repr=False)

    def create_queue(self, QueueName, Attributes):
        return self.aws.create_queue(QueueName, Attributes)

    def delete_queue(self, QueueUrl):
        return self.aws.delete_queue(QueueUrl)

    def list_queues(self, QueueNamePrefix=""):
        return {
            "QueueUrls": [
                queue.url
                for queue in self.aws.queues
                if queue.name.startswith(QueueNamePrefix)
            ]
        }


@attr.s
class ServiceResource(object):

    aws: MemoryAWS = attr.ib(repr=False)

    def create_queue(self, QueueName, Attributes):
        return self.aws.create_queue(QueueName, Attributes)

    def get_queue_by_name(self, QueueName):
        for queue in self.aws.queues:
            if queue.name == QueueName:
                return queue
        return None


@attr.s
class MemoryQueue(object):
    """
    In-memory queue which mimics the subset of SQS Queue object.

    Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/
         services/sqs.html#queue
    """

    aws: MemoryAWS = attr.ib()
    name: str = attr.ib()
    attributes: Dict[str, Dict[str, str]] = attr.ib()
    messages: List["MemoryMessage"] = attr.ib(factory=list)

    def __attrs_post_init__(self):
        self.attributes["QueueArn"] = self.name

    @property
    def url(self):
        return f"memory://{self.name}"

    def send_message(self, **kwargs):
        message = MemoryMessage.from_kwargs(self, kwargs)
        self.messages.append(message)
        return {"MessageId": message.message_id, "SequenceNumber": 0}

    def send_messages(self, Entries):
        res = []
        for message in Entries:
            res.append(self.send_message(**message))
        return {"Successful": res, "Failed": []}

    def receive_messages(self, WaitTimeSeconds="0", MaxNumberOfMessages="10", **kwargs):
        """
        Helper function which returns at most max_messages from the
        queue. Used in an infinite loop inside `get_raw_messages`
        """
        wait_seconds = int(WaitTimeSeconds)
        max_messages = int(MaxNumberOfMessages)

        ready_messages = []
        push_back_messages = []

        threshold = datetime.datetime.utcnow() + datetime.timedelta(
            seconds=wait_seconds
        )
        for message in self.messages:
            if message.execute_at > threshold or len(ready_messages) >= max_messages:
                push_back_messages.append(message)
            else:
                ready_messages.append(message)
        self.messages[:] = push_back_messages
        return ready_messages

    def delete_messages(self, Entries):
        """
        Delete messages implementation.

        See: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/
             services/sqs.html#SQS.Queue.delete_messages
        """
        message_ids = {entry["Id"] for entry in Entries}

        successfully_deleted = set()
        push_back_messages = []

        for message in self.messages:
            if message.message_id in message_ids:
                successfully_deleted.add(message.message_id)
            else:
                push_back_messages.append(message)
        self.messages[:] = push_back_messages

        didnt_deleted = message_ids.difference(successfully_deleted)
        return {
            "Successful": [{"Id": _id} for _id in successfully_deleted],
            "Failed": [{"Id": _id} for _id in didnt_deleted],
        }

    def delete(self):
        return self.aws.delete_queue(self.url)

    def __dict__(self):
        return {"QueueUrl": self.url}

    def __getitem__(self, item):
        return self.__dict__()[item]


@attr.s(frozen=True)
class MemoryMessage(object):
    """
    A mock class to mimic the AWS message

    Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/
         services/sqs.html#SQS.Message
    """

    queue_impl: MemoryQueue = attr.ib()

    # The message's contents (not URL-encoded).
    body: bytes = attr.ib()

    # Each message attribute consists of a Name, Type, and Value.
    message_attributes: Dict[str, Dict[str, str]] = attr.ib(factory=dict)

    # A map of the attributes requested in `` ReceiveMessage `` to their
    # respective values.
    attributes: Dict[str, Any] = attr.ib(factory=dict)

    # Internal attribute which contains the execution time.
    execute_at: datetime.datetime = attr.ib(factory=datetime.datetime.utcnow)

    # A unique identifier for the message
    message_id: str = attr.ib(factory=lambda: uuid.uuid4().hex)

    # The Message's receipt_handle identifier
    receipt_handle: str = attr.ib(factory=lambda: uuid.uuid4().hex)

    @classmethod
    def from_kwargs(cls, queue_impl, kwargs):
        """
        Make a message from kwargs, as provided by queue.send_message():

        Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/
             services/sqs.html#SQS.Queue.send_message
        """
        # required attributes
        body = kwargs["MessageBody"]
        message_atttributes = kwargs["MessageAttributes"]

        # optional attributes
        attributes = {"ApproximateReceiveCount": 1}
        if "MessageDeduplicationId" in kwargs:
            attributes["MessageDeduplicationId"] = kwargs["MessageDeduplicationId"]

        if "MessageGroupId" in kwargs:
            attributes["MessageGroupId"] = kwargs["MessageGroupId"]

        execute_at = datetime.datetime.utcnow()
        if "DelaySeconds" in kwargs:
            delay_seconds_int = int(kwargs["DelaySeconds"])
            execute_at += datetime.timedelta(seconds=delay_seconds_int)

        return MemoryMessage(
            queue_impl, body, message_atttributes, attributes, execute_at
        )

    def change_visibility(self, VisibilityTimeout="0", **kwargs):
        timeout = int(VisibilityTimeout)
        execute_at = datetime.datetime.utcnow() + datetime.timedelta(seconds=timeout)
        message = attr.evolve(self, execute_at=execute_at)
        message.attributes["ApproximateReceiveCount"] += 1
        self.queue_impl.messages.append(message)



import importlib
import logging
from inspect import Signature
from typing import Any

logger = logging.getLogger(__name__)


def _bind_args(callback, args, kwargs, drop_extra):
    sig = Signature.from_callable(callback)
    bind_args = list(args)
    bind_kwargs = {ensure_string(k): v for k, v in kwargs.items()}

    if drop_extra:
        # Compatibility with Werkzeug's `drop_extra`: if true, we drop extra
        # positional and keyword arguments.
        # This implementation is not perfect, but should be good enough for most
        # real-world use cases.
        extra_args, extra_kwargs = [], {}

        # - Is there a *args argument?
        if not any(
            param.kind == param.VAR_POSITIONAL for param in sig.parameters.values()
        ):
            # No: drop extra positional arguments
            all_pos = [
                param.name
                for param in sig.parameters.values()
                if param.kind in (param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD)
            ]
            nb_pos = len(all_pos)
            bind_args, extra_args = bind_args[:nb_pos], bind_args[nb_pos:]

        # - Is there a **kwargs argument?
        if not any(
            param.kind == param.VAR_KEYWORD for param in sig.parameters.values()
        ):
            # No: drop extra keyword arguments
            all_kw = {
                param.name
                for param in sig.parameters.values()
                if param.kind in (param.POSITIONAL_OR_KEYWORD, param.KEYWORD_ONLY)
            }
            new_bind_kwargs = {}
            for key, value in bind_kwargs.items():
                if key in all_kw:
                    new_bind_kwargs[key] = value
                else:
                    extra_kwargs[key] = value
            bind_kwargs = new_bind_kwargs

        if extra_args or extra_kwargs:
            logger.debug(
                "Dropped %d extra positional arguments and %d extra keyword arguments",
                len(extra_args),
                len(extra_kwargs),
                extra={
                    "extra_args": extra_args,
                    "extra_kwargs": extra_kwargs,
                },
            )

    bound_args = sig.bind(*bind_args, **bind_kwargs)
    return bound_args


def validate_arguments(callback, args, kwargs, drop_extra=True):
    """Checks if the function accepts the provided arguments and keyword
    arguments. Returns a new `(args, kwargs)` tuple that can be passed to the
    function without causing a TypeError due to an incompatible signature.

    If the arguments are invalid, a TypeError is raised.

    Similar to Werkzeug's old `validate_arguments` function, but doesn't modify
    passed args and kwargs.
    """
    bound_args = _bind_args(callback, args, kwargs, drop_extra)
    return (bound_args.args, bound_args.kwargs)


def bind_arguments(callback, args, kwargs):
    """Bind the arguments provided into a dict, returning a dict of bound
    keyword arguments.

    Similar to Werkzeug's old `bind_arguments` function, but doesn't modify
    passed args and kwargs.
    """
    bound_args = _bind_args(callback, args, kwargs, False)
    return bound_args.arguments


def string_to_object(string: str) -> Any:
    """
    Convert full path string representation of the object to object itself.
    """
    chunks = string.rsplit(".", 1)
    if len(chunks) != 2:
        raise RuntimeError(
            "{} doesn't represent a full module path to object".format(string)
        )
    module_name, object_name = chunks
    mod = importlib.import_module(module_name)
    return getattr(mod, object_name)


def instantiate_from_string(string, **init_kwargs):
    """
    Create an object from a classname string and init kwargs.
    """
    class_ = string_to_object(string)
    return class_(**init_kwargs)


def instantiate_from_dict(options, maker_key="maker", **extra_init_kwargs):
    """
    Create an object from options with classname and init kwargs.
    """
    kwargs = options.copy()
    classname_value = kwargs.pop(maker_key)
    init_kwargs = dict(kwargs, **extra_init_kwargs)
    return string_to_object(classname_value)(**init_kwargs)


def ensure_string(obj: Any, encoding="utf-8", errors="strict") -> str:
    """Make sure an object is converted to a proper string representation."""
    if isinstance(obj, str):
        return obj
    elif isinstance(obj, bytes):
        return obj.decode(encoding, errors)
    else:
        return str(obj)



from typing import Any, Dict, Optional

import attr

from sqs_workers.utils import (
    instantiate_from_dict,
    instantiate_from_string,
    string_to_object,
)


@attr.s(frozen=True)
class Config(object):
    """
    Config object with hierarchy support.
    """

    parent: Optional["Config"] = attr.ib(repr=False, default=None)
    options: Dict[str, Any] = attr.ib(factory=dict)
    maker_key = attr.ib(default="maker")

    def __setitem__(self, key, value):
        self.options.__setitem__(key, value)

    def __getitem__(self, item):
        if item in self.options:
            return self.options[item]
        if self.parent:
            return self.parent[item]
        raise KeyError("{0} is undefined".format(item))

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            return default

    def get_object(self, item):
        """
        Get an object (usually a class) from the config.
        """
        value = self[item]
        return string_to_object(value)

    def get_instance(self, item, **kwargs):
        """
        Get an instances form the config and optional set of kwargs.
        """
        value = self[item]
        if isinstance(value, dict):
            return instantiate_from_dict(value, maker_key=self.maker_key, **kwargs)
        else:
            return instantiate_from_string(value, **kwargs)

    def make_child(self, options=None):
        if options is None:
            options = {}
        return Config(parent=self, options=options)



import logging
from functools import partial
from typing import TYPE_CHECKING, Callable, Optional

import attr

from sqs_workers import codecs
from sqs_workers.context import SQSContext
from sqs_workers.utils import validate_arguments

if TYPE_CHECKING:
    from sqs_workers.queue import GenericQueue

logger = logging.getLogger(__name__)

DEFAULT_CONTEXT_VAR = "context"


@attr.s
class Processor(object):
    """
    Processor which calls its function for each incoming message.

    fn() accepts a single parameter "message" which is a decoded body of an
    SQS message
    """

    queue: "GenericQueue" = attr.ib()
    fn: Optional[Callable] = attr.ib(default=None)
    job_name: str = attr.ib(default="")
    pass_context: bool = attr.ib(default=False)
    context_var: str = attr.ib(default=DEFAULT_CONTEXT_VAR)

    @classmethod
    def maker(cls, **kwargs):
        return partial(cls, **kwargs)

    def process_message(self, message):
        """
        Takes an individual message and sends it to the decorated call handler function.
        """
        extra = {
            "message_id": message.message_id,
            "queue_name": self.queue.name,
            "job_name": self.job_name,
        }
        logger.debug("Process %s.%s", self.queue.name, self.job_name, extra=extra)

        try:
            content_type, job_kwargs, job_context = self.deserialize_message(message)
            extra["job_content_type"] = content_type
            self.process(job_kwargs, job_context)
        except Exception:
            logger.exception(
                "Error while processing %s.%s",
                self.queue.name,
                self.job_name,
                extra=extra,
            )
            return False
        else:
            return True

    def process_messages(self, messages):
        """
        Take a list of messages and send them in a batch
        to the decorated call handler function.
        """
        message_ids = [m.message_id for m in messages]
        extra = {
            "message_ids": message_ids,
            "queue_name": self.queue.name,
            "job_name": self.job_name,
        }
        logger.debug(
            "Processing batch for %s.%s",
            self.queue.name,
            self.job_name,
            extra=extra,
        )

        try:
            self.process_batch(messages)
        except Exception:
            logger.exception(
                "Error while processing batch for %s.%s",
                self.queue.name,
                self.job_name,
                extra=extra,
            )
            return False
        else:
            return True

    def process(self, job_kwargs, job_context):
        effective_kwargs = job_kwargs.copy()
        if self.pass_context:
            effective_kwargs[self.context_var] = job_context
        return call_handler(self.fn, [], effective_kwargs)

    def process_batch(self, messages):
        deserialized_messages = []
        for message in messages:
            _, job_kwargs, job_context = self.deserialize_message(message)
            if self.pass_context:
                job_kwargs[self.context_var] = job_context
            deserialized_messages.append(job_kwargs)

        call_handler(self.fn, [deserialized_messages], {})

    def copy(self, **kwargs):
        """
        Create a new instance of the processor, optionally updating
        arguments of the constructor from update_kwargs
        """
        return attr.evolve(self, **kwargs)

    @staticmethod
    def deserialize_message(message):
        content_type = get_job_content_type(message)
        codec = codecs.get_codec(content_type)
        job_kwargs = codec.deserialize(message.body)
        job_context = get_job_context(message, codec)
        return content_type, job_kwargs, job_context


def get_job_content_type(job_message):
    attrs = job_message.message_attributes or {}
    return (attrs.get("ContentType") or {}).get("StringValue")


def get_job_context(job_message, codec):
    attrs = job_message.message_attributes or {}
    serialized = (attrs.get("JobContext") or {}).get("StringValue")
    if not serialized:
        return SQSContext()
    deserialized = codec.deserialize(serialized)
    return SQSContext.from_dict(deserialized)


def call_handler(fn, args, kwargs):
    try:
        handler_args, handler_kwargs = validate_arguments(fn, args, kwargs)
    except TypeError:
        # it may happen, if "fn" is not a function (but
        # a mock object, for example)
        handler_args, handler_kwargs = args, kwargs
    return fn(*handler_args, **handler_kwargs)



import logging
import uuid
from collections import defaultdict
from contextlib import contextmanager
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional

import attr

from sqs_workers import DEFAULT_BACKOFF, codecs
from sqs_workers.async_task import AsyncTask
from sqs_workers.backoff_policies import BackoffPolicy
from sqs_workers.batching import BatchingConfiguration, NoBatching
from sqs_workers.core import BatchProcessingResult, get_job_name
from sqs_workers.exceptions import SQSError
from sqs_workers.processors import DEFAULT_CONTEXT_VAR, Processor
from sqs_workers.shutdown_policies import NEVER_SHUTDOWN

DEFAULT_MESSAGE_GROUP_ID = "default"
SEND_BATCH_SIZE = 10

if TYPE_CHECKING:
    from sqs_workers import SQSEnv

logger = logging.getLogger(__name__)


class SQSBatchError(SQSError):
    def __init__(self, errors):
        self.errors = errors


@attr.s
class GenericQueue(object):
    env: "SQSEnv" = attr.ib(repr=False)
    name: str = attr.ib()
    backoff_policy: BackoffPolicy = attr.ib(default=DEFAULT_BACKOFF)
    batching_policy: BatchingConfiguration = attr.ib(default=NoBatching())
    _queue = attr.ib(repr=False, default=None)

    @classmethod
    def maker(cls, **kwargs):
        return partial(cls, **kwargs)

    def process_queue(self, shutdown_policy=NEVER_SHUTDOWN, wait_second=10):
        """
        Run worker to process one queue in the infinite loop
        """
        logger.debug(
            "Start processing queue %s",
            self.name,
            extra={
                "queue_name": self.name,
                "wait_seconds": wait_second,
                "shutdown_policy": repr(shutdown_policy),
            },
        )
        while True:
            result = self.process_batch(wait_seconds=wait_second)
            shutdown_policy.update_state(result)
            if shutdown_policy.need_shutdown():
                logger.debug(
                    "Stop processing queue %s",
                    self.name,
                    extra={
                        "queue_name": self.name,
                        "wait_seconds": wait_second,
                        "shutdown_policy": repr(shutdown_policy),
                    },
                )
                break

    def process_batch(self, wait_seconds=0) -> BatchProcessingResult:
        """
        Process a batch of messages from the queue (10 messages at most), return
        the number of successfully processed messages, and exit
        """
        queue = self.get_queue()

        if self.batching_policy.batching_enabled:
            return self._process_messages_in_batch(queue, wait_seconds)

        return self._process_messages_individually(queue, wait_seconds)

    def _process_messages_in_batch(self, queue, wait_seconds):
        messages = self.get_raw_messages(wait_seconds, self.batching_policy.batch_size)
        result = BatchProcessingResult(self.name)

        success = self.process_messages(messages)

        for message in messages:
            result.update_with_message(message, success)
            if success:
                entry = {
                    "Id": message.message_id,
                    "ReceiptHandle": message.receipt_handle,
                }
                queue.delete_messages(Entries=[entry])
            else:
                timeout = self.backoff_policy.get_visibility_timeout(message)
                message.change_visibility(VisibilityTimeout=timeout)
        return result

    def _process_messages_individually(self, queue, wait_seconds):
        messages = self.get_raw_messages(wait_seconds)
        result = BatchProcessingResult(self.name)

        for message in messages:
            success = self.process_message(message)
            result.update_with_message(message, success)
            if success:
                entry = {
                    "Id": message.message_id,
                    "ReceiptHandle": message.receipt_handle,
                }
                queue.delete_messages(Entries=[entry])
            else:
                timeout = self.backoff_policy.get_visibility_timeout(message)
                message.change_visibility(VisibilityTimeout=timeout)
        return result

    def process_message(self, message: Any) -> bool:
        """
        Process single message.

        Return True if processing went successful
        """
        raise NotImplementedError()

    def process_messages(self, messages: List[Any]) -> bool:
        """
        Process a batch of messages.

        Return True if processing went successful
        for every message in the batch
        """
        raise NotImplementedError()

    def get_raw_messages(self, wait_seconds, max_messages=10):
        """
        Return raw messages from the queue, addressed by its name
        """
        kwargs = {
            "WaitTimeSeconds": wait_seconds,
            "MaxNumberOfMessages": max_messages if max_messages <= 10 else 10,
            "MessageAttributeNames": ["All"],
            "AttributeNames": ["All"],
        }
        queue = self.get_queue()

        if max_messages <= 10:
            return queue.receive_messages(**kwargs)

        messages_left_on_queue = True
        received_messages = []

        # receive_messages will only return a maximum batch of 10 messages per call
        # if the client requests more than that we must poll multiple times
        while messages_left_on_queue and len(received_messages) < max_messages:
            messages = queue.receive_messages(**kwargs)
            received_messages.extend(messages)
            if len(messages) < 10:
                messages_left_on_queue = False

        return received_messages

    def drain_queue(self, wait_seconds=0):
        """
        Delete all messages from the queue without calling purge().
        """
        queue = self.get_queue()
        deleted_count = 0
        while True:
            messages = self.get_raw_messages(wait_seconds)
            if not messages:
                break
            entries = [
                {"Id": msg.message_id, "ReceiptHandle": msg.receipt_handle}
                for msg in messages
            ]
            queue.delete_messages(Entries=entries)
            deleted_count += len(messages)
        return deleted_count

    def get_sqs_queue_name(self):
        """
        Take "high-level" (user-visible) queue name and return SQS
        ("low level") name by simply prefixing it. Used to create namespaces
        for different environments (development, staging, production, etc)
        """
        return self.env.get_sqs_queue_name(self.name)

    def get_queue(self):
        """
        Helper function to return queue object.
        """
        if self._queue is None:
            self._queue = self.env.sqs_resource.get_queue_by_name(
                QueueName=self.get_sqs_queue_name()
            )
        return self._queue


@attr.s
class RawQueue(GenericQueue):
    processor: Optional[Callable] = attr.ib(default=None)

    def raw_processor(self):
        """
        Decorator to assign processor to handle jobs from the raw queue

        Usage example:

            cron = sqs.queue('cron')

            @cron.raw_processor()
            def process(message):
                print(message.body)
        """

        def func(processor):
            return self.connect_raw_processor(processor)

        return func

    def connect_raw_processor(self, processor):
        """
        Assign raw processor (a function) to handle jobs.
        """
        extra = {
            "queue_name": self.name,
            "processor_name": processor.__module__ + "." + processor.__name__,
        }
        logger.debug(
            "Connect %s to processor %s",
            extra["queue_name"],
            extra["processor_name"],
            extra=extra,
        )
        self.processor = processor

    def add_raw_job(
        self,
        message_body: str,
        delay_seconds: int = 0,
        deduplication_id: Optional[str] = None,
        group_id: str = DEFAULT_MESSAGE_GROUP_ID,
    ):
        """
        Add raw message to the queue
        """

        kwargs = {
            "MessageBody": message_body,
            "DelaySeconds": delay_seconds,
            "MessageAttributes": {},
        }
        if self.name.endswith(".fifo"):
            # if queue name ends with .fifo, then according to the AWS specs,
            # it's a FIFO queue, and requires group_id.
            # Otherwise group_id can be set to None
            if group_id is None:
                group_id = DEFAULT_MESSAGE_GROUP_ID
            if deduplication_id is not None:
                kwargs["MessageDeduplicationId"] = str(deduplication_id)
            if group_id is not None:
                kwargs["MessageGroupId"] = str(group_id)

        ret = self.get_queue().send_message(**kwargs)
        return ret["MessageId"]

    def process_message(self, message: Any) -> bool:
        """
        Sends a single message to the call handler.

        Return True if processing went successful
        """
        extra = {"message_id": message.message_id, "queue_name": self.name}
        if not self.processor:
            logger.warning("No processor set for %s", extra["queue_name"], extra=extra)
            return False

        logger.debug(
            "Process %s.%s", extra["queue_name"], extra["message_id"], extra=extra
        )

        try:
            self.processor(message)
        except Exception:
            logger.exception(
                "Error while processing %s.%s",
                extra["queue_name"],
                extra["message_id"],
                extra=extra,
            )
            return False
        else:
            return True

    def process_messages(self, messages: List[Any]) -> bool:
        """
        Sends a list of messages to the call handler

        Return True if processing went successful for all messages in the batch.
        """
        message_ids = [m.message_id for m in messages]
        extra = {"message_ids": message_ids, "queue_name": self.name}
        if not self.processor:
            logger.warning(
                "No processor set for %s",
                extra["queue_name"],
                extra=extra,
            )
            return False

        logger.debug(
            "Process %s for ids %s",
            extra["queue_name"],
            extra["message_ids"],
            extra=extra,
        )

        try:
            self.processor(messages)
        except Exception:
            logger.exception(
                "Error while processing %s for ids %s",
                extra["queue_name"],
                extra["message_ids"],
                extra=extra,
            )
            return False
        else:
            return True


@attr.s
class JobQueue(GenericQueue):
    processors: Dict[str, Processor] = attr.ib(factory=dict)

    _batch_level: int = attr.ib(default=0, repr=False)
    _batched_messages: List[Dict] = attr.ib(factory=list, repr=False)

    def processor(self, job_name, pass_context=False, context_var=DEFAULT_CONTEXT_VAR):
        """
        Decorator to assign processor to handle jobs with the name job_name
        from the queue queue_name

        Usage example:

            queue = sqs.queue('q1')

            @queue.processor('say_hello')
            def say_hello(name):
                print("Hello, " + name)

        Then you can add messages to the queue by calling ".delay" attribute
        of a newly created object:

            say_hello.delay(name='John')

        Here, instead of calling function locally, it will be pushed to the
        queue (essentially, mimics the non-blocking call of the function).

        If you still want to call function locally, you can call

            say_hello(name='John')
        """

        def fn(processor):
            return self.connect_processor(
                job_name, processor, pass_context, context_var
            )

        return fn

    def connect_processor(
        self, job_name, processor, pass_context=False, context_var=DEFAULT_CONTEXT_VAR
    ):
        """
        Assign processor (a function) to handle jobs with the name job_name
        from the queue queue_name
        """
        extra = {
            "queue_name": self.name,
            "job_name": job_name,
            "processor_name": processor.__module__ + "." + processor.__name__,
        }
        logger.debug(
            "Connect %s.%s to processor %s",
            extra["queue_name"],
            extra["job_name"],
            extra["processor_name"],
            extra=extra,
        )
        self.processors[job_name] = self.env.processor_maker(
            queue=self,
            fn=processor,
            pass_context=pass_context,
            context_var=context_var,
            job_name=job_name,
        )
        return AsyncTask(self, job_name, processor)

    def get_processor(self, job_name):
        """
        Helper function to return a processor for the queue
        """
        return self.processors.get(job_name)

    def open_add_batch(self):
        """
        Open an add batch.
        """
        self._batch_level += 1

    def close_add_batch(self):
        """
        Close an add batch.
        """
        self._batch_level = max(0, self._batch_level - 1)
        self._flush_batch_if_needed()

    @contextmanager
    def add_batch(self):
        """
        Context manager to add jobs in batch.

        Inside this context manager, jobs won't be added to the queue
        immediately, but grouped by batches of 10. Once open, new jobs won't be
        sent to SQS immediately, but added to a local cache. Then, once this
        cache is big enough or when closing the batch, all messages will be
        sent, using as few requests as possible.
        """
        self.open_add_batch()
        try:
            yield
        finally:
            self.close_add_batch()

    def _should_flush_batch(self):
        """
        Check if a message batch should be flushed, i.e. all messages should be sent
        and removed from the internal cache.

        Right now this only checks the number of messages. In the future we may
        want to improve that by also ensuring the batch size is small enough.
        """
        max_size = SEND_BATCH_SIZE if self._batch_level > 0 else 1
        return len(self._batched_messages) >= max_size

    def _flush_batch_if_needed(self):
        queue = self.get_queue()

        # There should be at most 1 batch to send. But just in case, prepare to
        # send more than that.
        while self._should_flush_batch():
            msgs = self._batched_messages[:SEND_BATCH_SIZE]
            self._batched_messages = self._batched_messages[SEND_BATCH_SIZE:]

            # Add Ids
            msg_by_id = {}
            for msg in msgs:
                id_ = uuid.uuid4().hex
                msg["Id"] = id_
                msg_by_id[id_] = msg

            res = queue.send_messages(Entries=msgs)

            # Handle errors
            if res.get("Failed", []):
                errors = res["Failed"]
                for error in errors:
                    error["_message"] = msg_by_id[error["Id"]]
                raise SQSBatchError(errors)

    def add_job(
        self,
        job_name,
        _content_type=codecs.DEFAULT_CONTENT_TYPE,
        _delay_seconds=None,
        _deduplication_id=None,
        _group_id=None,
        **job_kwargs
    ):
        """
        Add job to the queue. The body of the job will be converted to the text
        with one of the codes (by default it's "pickle")
        """
        codec = codecs.get_codec(_content_type)
        message_body = codec.serialize(job_kwargs)
        job_context = codec.serialize(self.env.context.to_dict())
        return self.add_raw_job(
            job_name,
            message_body,
            job_context,
            _content_type,
            _delay_seconds,
            _deduplication_id,
            _group_id,
        )

    def add_raw_job(
        self,
        job_name,
        message_body,
        job_context,
        content_type,
        delay_seconds,
        deduplication_id,
        group_id,
    ):
        """
        Low-level function to put message to the queue
        """
        # if queue name ends with .fifo, then according to the AWS specs,
        # it's a FIFO queue, and requires group_id.
        # Otherwise group_id can be set to None
        if group_id is None and self.name.endswith(".fifo"):
            group_id = DEFAULT_MESSAGE_GROUP_ID

        queue = self.get_queue()
        kwargs = {
            "MessageBody": message_body,
            "MessageAttributes": {
                "ContentType": {"StringValue": content_type, "DataType": "String"},
                "JobContext": {"StringValue": job_context, "DataType": "String"},
                "JobName": {"StringValue": job_name, "DataType": "String"},
            },
        }
        if delay_seconds is not None:
            kwargs["DelaySeconds"] = int(delay_seconds)
        if deduplication_id is not None:
            kwargs["MessageDeduplicationId"] = str(deduplication_id)
        if group_id is not None:
            kwargs["MessageGroupId"] = str(group_id)

        if self._batch_level > 0:
            self._batched_messages.append(kwargs)
            self._flush_batch_if_needed()
            return None
        else:
            ret = queue.send_message(**kwargs)
            return ret["MessageId"]

    def process_message(self, message: Any) -> bool:
        """
        Sends a single message to the call handler.

        Return True if processing went successful
        """
        job_name = get_job_name(message)
        processor = self.get_processor(job_name)
        if processor:
            return processor.process_message(message)
        else:
            return self.process_message_fallback(job_name)

    def process_messages(self, messages: List[Any]) -> bool:
        """
        Sends a list of messages to the call handler

        Return True if processing went successful for all messages in the batch.
        """
        messages_by_job_name = defaultdict(list)
        for message in messages:
            job_name = get_job_name(message)
            messages_by_job_name[job_name].append(message)

        results = []

        for job_name, grouped_messages in messages_by_job_name.items():
            processor = self.get_processor(job_name)
            if processor:
                result = processor.process_messages(grouped_messages)
                results.append(result)
            else:
                result = self.process_message_fallback(job_name)
                results.append(result)

        return all(results)

    def process_message_fallback(self, job_name):
        logger.warning("Error while processing %s.%s", self.name, job_name)
        return False

    def copy_processors(self, dst_queue: "JobQueue"):
        """
        Copy processors from self to dst_queue. Can be helpful to process d
        ead-letter queue with processors from the main queue.

        Usage example.

        sqs = SQSEnv()
        ...
        foo = sqs.queue('foo')
        foo_dead = sqs.queue('foo_dead')
        foo.copy_processors('foo_dead')
        foo_dead.process_queue(shutdown_policy=IdleShutdown(10))

        Here the queue "foo_dead" will be processed with processors from the
        queue "foo".
        """
        for job_name, processor in self.processors.items():
            dst_queue.processors[job_name] = processor.copy(queue=dst_queue)



import base64
import json
import pickle
import zlib

DEFAULT_CONTENT_TYPE = "pickle"

# Highest version of the protocol, understood by Python2.
PICKLE_PY2_COMPAT_PROTO = 2


class JSONCodec(object):
    @staticmethod
    def serialize(message):
        return json.dumps(message)

    @staticmethod
    def deserialize(serialized):
        return json.loads(serialized)


class PickleCodec(object):
    @staticmethod
    def serialize(message):
        binary_data = pickle.dumps(message, protocol=PICKLE_PY2_COMPAT_PROTO)
        compressed_data = zlib.compress(binary_data)
        return base64.urlsafe_b64encode(compressed_data).decode("latin1")

    @staticmethod
    def deserialize(serialized):
        compressed_data = base64.urlsafe_b64decode(serialized.encode("latin1"))
        binary_data = zlib.decompress(compressed_data)
        return pickle.loads(binary_data)


def get_codec(content_type):
    return CONTENT_TYPES_CODECS[content_type]


CONTENT_TYPES_CODECS = {"json": JSONCodec, "pickle": PickleCodec}



from abc import ABC, abstractmethod


class BatchingConfiguration(ABC):
    """
    Defining the contract for configuring whether a processor sends
    messages 1 by 1 to a call handler, or as a list of messages.
    """

    @property
    @abstractmethod
    def batching_enabled(self) -> bool:
        """
        If false, messages are sent 1 by 1 to the call handler
        If true, messages are sent as a list to the call handler
        """
        ...

    @property
    @abstractmethod
    def batch_size(self) -> int:
        """
        Number of messages to process at once if batching_enabled
        """
        ...


class NoBatching(BatchingConfiguration):
    """Configures the processor to send messages 1 by 1 to the call handler"""

    @property
    def batching_enabled(self) -> bool:
        return False

    @property
    def batch_size(self) -> int:
        return 1


class BatchMessages(BatchingConfiguration):
    """Configures the processor to send a list of messages to the call handler"""

    def __init__(self, batch_size):
        self.number_of_messages = batch_size

    @property
    def batching_enabled(self) -> bool:
        return True

    @property
    def batch_size(self) -> int:
        return self.number_of_messages



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sqs_workers']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=22.1.0,<23.0.0', 'boto3>=1.24.62,<2.0.0']

setup_kwargs = {
    'name': 'sqs-workers',
    'version': '0.5.8',
    'description': 'An opinionated queue processor for Amazon SQS',
    'long_description': '# SQS Workers\n\nAn opinionated queue processor for Amazon SQS.\n\n## Usage\n\nInstall the package with\n\n```bash\npip install sqs-workers\n```\n\nConfigure your boto3 library to provide access requisites for your installation with [something like this](https://boto3.readthedocs.io/en/latest/guide/quickstart.html#configuration):\n\n```bash\naws configure\n```\n\nDon\'t forget to set your preferred AWS region.\n\nThen you will start managing two systems (most likely, from the same codebase): one of them adds messages to the queue, and another one executes them.\n\n```python\nfrom sqs_workers import SQSEnv, create_standard_queue\n\n# This environment will use AWS requisites from ~/.aws/ directory\nsqs = SQSEnv()\n\n# Create a new queue.\n# Note that you can use the AWS web interface for the same action as well, the\n# web interface provides more options. You only need to do it once.\ncreate_standard_queue(sqs, "emails")\n\n# Get the queue object\nqueue = sqs.queue("emails")\n\n# Register a queue processor\n@queue.processor("send_email")\ndef send_email(to, subject, body):\n    print(f"Sending email {subject} to {to}")\n```\n\nThen there are two ways of adding tasks to the queue. Classic (aka "explicit"):\n\n```python\nqueue.add_job("send_email", to="user@example.com", subject="Hello world", body="hello world")\n```\n\nAnd the "Celery way" (we mimic the Celery API to some extent)\n\n```python\nsend_email.delay(to="user@example.com", subject="Hello world", body="hello world")\n```\n\nTo process the queue, you have to run workers manually. Create a new file which will contain the definition of the sqs object and register all processors (most likely, by importing necessary modules from your project), and then run SQS\n\n```python\nfrom sqs_workers import SQSEnv\nsqs = SQSEnv()\n...\nsqs.queue(\'emails\').process_queue()\n```\n\nIn production, we usually don\'t handle multiple queues in the same process, but for the development environment it\'s easier to tackle with all the queues at once with\n\n```python\nsqs.process_queues()\n```\n\n## Serialization\n\nThere are two serializers: JSON and pickle.\n\n## Baked tasks\n\nYou can create so-called "baked async tasks," entities which, besides the task itself, contain arguments which have to be used to call the task.\n\nThink of baked tasks as of light version of [Celery signatures](http://docs.celeryproject.org/en/latest/userguide/canvas.html#signatures)\n\n```python\ntask = send_email.bake(to=\'user@example.com\', subject=\'Hello world\', body=\'hello world\')\ntask.delay()\n```\n\nIs the same as\n\n```python\nsend_email.delay(to=\'user@example.com\', subject=\'Hello world\', body=\'hello world\')\n```\n\n## Batch Writes\n\nIf you have many tasks to enqueue, it may be more efficient to use batching when adding them:\n\n```python\n# Classic ("explicit") API\nwith queue.add_batch():\n    queue.add_job("send_email", to="user1@example.com", subject="Hello world", body="hello world")\n    queue.add_job("send_email", to="user2@example.com", subject="Hello world", body="hello world")\n\n# "Celery way"\nwith send_email.batch():\n    send_email.delay(to="user1@example.com", subject="Hello world", body="hello world")\n    send_email.delay(to="user2@example.com", subject="Hello world", body="hello world")\n```\n\nWhen batching is enabled:\n\n- tasks are added to SQS by batches of 10, reducing the number of AWS operations\n- it is not possible to get the task `MessageId`, as it is not known until the batch is sent\n- care has to be taken about message size, as SQS limits both the individual message size and the maximum total payload size to 256 kB.\n\n## Batch Reads\n\nIf you wish to consume and process batches of messages from a queue at once (say to speed up ingestion)\nyou can set the `batching_policy` at queue level.\n\nThe underlying function is expected to have a single parameter which will receive the list of messages.\n\n```python\nfrom sqs_workers.batching import BatchMessages\nfrom sqs_workers import SQSEnv, create_standard_queue\n\nsqs = SQSEnv()\n\ncreate_standard_queue(sqs, "emails")\n\nqueue = sqs.queue("emails", batching_policy=BatchMessages(batch_size=10))\n\n@queue.processor("send_email")\ndef send_email(messages: list):\n    for email in messages:\n        print(f"Sending email {email[\'subject\']} to {email[\'to\']}")\n```\n\nThis function will receive up to 10 messages at once based on:\n\n* How many are available on the queue being consumed\n* How many of those messages on the `emails` queue are for the `send_email` job\n\n## Synchronous task execution\n\nIn Celery, you can run any task synchronously if you just call it as a function with arguments. Our AsyncTask raises a RuntimeError for this case.\n\n```python\nsend_email(to=\'user@example.com\', subject=\'Hello world\', body=\'hello world\')\n...\nRuntimeError: Async task email.send_email called synchronously (probably,\nby mistake). Use either AsyncTask.run(...) to run the task synchronously\nor AsyncTask.delay(...) to add it to the queue.\n```\n\nIf you want to run a task synchronously, use `run()` method of the task.\n\n```\nsend_email.run(to=\'user@example.com\', subject=\'Hello world\', body=\'hello world\')\n```\n\n## FIFO queues\n\nFifo queues can be created with `create_fifo_queue` and have to have the name, which ends with ".fifo".\n\n```python\nfrom sqs_workers import SQSEnv, create_fifo_queue\nsqs = SQSEnv()\n\ncreate_fifo_queue(sqs, \'emails_dead.fifo\')\ncreate_fifo_queue(sqs, \'emails.fifo\',\n    redrive_policy=sqs.redrive_policy(\'emails_dead.fifo\', 3)\n)\n```\n\nUnless the flag `content_based_deduplication` is set, every message has to be sent with an attribute `_deduplication_id`. By default, all messages have the same message group `default`, but you can change it with `_group_id`.\n\n```python\nsqs.queue("emails.fifo").add_job(\n    \'send_email\', _deduplication_id=subject, _group_id=email, **kwargs)\n```\n\n[More about FIFO queues on AWS](https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/FIFO-queues.html)\n\n## Exception processing\n\nIf task processing ended up with an exception, the error is logged, and the task is returned back to the queue after a while. The exact behavior is defined by queue settings.\n\n## Custom processors\n\nYou can define your own processor if you need to perform some specific actions before of after executing a particular task.\n\nExample for the custom processor\n\n```python\nfrom sqs_workers import SQSEnv\nfrom sqs_workers.processors import Processor\n\nclass CustomProcessor(Processor):\n    def process(self, job_kwargs):\n        print(f\'Processing {self.queue_name}.{self.job_name} with {job_kwargs}\')\n        super(CustomProcessor, self).process(job_kwargs)\n\nsqs = SQSEnv(processor_maker=CustomProcessor)\n```\n\n## Working with contexts\n\nContext is implicitly passed to the worker via the job message and can be used for logging or profiling purposes.\n\nUsage example.\n\n```python\nqueue = sqs.queue("q1")\n\n@queue.processor(\'q1\', \'hello_world\', pass_context=True)\ndef hello_world(username=None, context=None):\n    print(f\'Hello {username} from {context["remote_addr"]}\')\n\nwith sqs.context(remote_addr=\'127.0.0.1\'):\n    hello_world.delay(username=\'Foo\')\n\nqueue.process_batch()\n```\n\nAlternatively, you can set a context like this.\n\n```python\nsqs.context[\'remote_addr\'] = \'127.0.0.1\'\nhello_world.delay(username=\'Foo\')\n```\n\nAnd then, when the context needs to be cleared:\n\n```python\nsqs.context.clear()\n```\n\nIn a web application, you usually call it at the end of the processing of the web request.\n\n## Automatic applying of the context for all tasks\n\nInstead of dealing with the context inside every processing function, you can perform this in processors by subclassing them.\n\n```python\nimport os\nfrom sqs_workers import SQSEnv\nfrom sqs_workers.processors import Processor\n\nclass CustomProcessor(Processor):\n    def process(self, job_kwargs, job_context):\n        os.environ[\'REMOTE_ADDR\'] = job_context[\'remote_addr\']\n        super(CustomProcessor, self).process(job_kwargs, job_context)\n        os.unsetenv(\'REMOTE_ADDR\')\n\nsqs = SQSEnv(\n    processor_maker=CustomProcessor,\n)\n```\n\n## Raw queues\n\nRaw queues can have only one processor, which should be a function, accepting the message as its only argument.\n\nRaw queues are helpful to process messages, added to SQS from external sources, such as CloudWatch events.\n\nYou start the processor the same way, creating a new standard queue if needed.\n\n```python\nfrom sqs_workers import SQSEnv, create_standard_queue\nsqs = SQSEnv()\ncreate_standard_queue(sqs, \'cron\')\n```\n\nThen you get a queue, but provide a queue_maker parameter to it, to create a queue of the necessary type, and define a processor for it.\n\n```python\nfrom sqs_workers import RawQueue\n\ncron = sqs.queue(\'cron\', RawQueue)\n\n@cron.raw_processor()\ndef processor(message):\n    print(message.body)\n```\n\nThen start processing your queue as usual:\n\n```python\ncron.process_queue()\n```\n\nYou can also send raw messages to the queue, but this is probably less useful:\n\n```python\ncron.add_raw_job("Hello world")\n```\n\n## Processing Messages from CloudWatch\n\nBy default message body by CloudWatch scheduler has following JSON structure.\n\n```json\n{\n  "version": "0",\n  "id": "a9a10406-9a1f-0ddc-51ae-08db551fac42",\n  "detail-type": "Scheduled Event",\n  "source": "aws.events",\n  "account": "NNNNNNNNNN",\n  "time": "2019-09-20T09:19:56Z",\n  "region": "eu-west-1",\n  "resources": ["arn:aws:events:eu-west-1:NNNNNNNNNN:rule/Playground"],\n  "detail": {}\n}\n```\n\nHeaders of the message:\n\n```python\n{\n    \'SenderId\': \'AIDAJ2E....\',\n    \'ApproximateFirstReceiveTimestamp\': \'1568971264367\',\n    \'ApproximateReceiveCount\': \'1\',\n    \'SentTimestamp\': \'1568971244845\',\n}\n```\n\nYou can pass any valid JSON as a message, though, and it will be passed as-is to the message body. Something like this:\n\n```json\n{ "message": "Hello world" }\n```\n\n## Dead-letter queues and redrive\n\nOn creating the queue, you can set the fallback dead-letter queue and redrive policy, which can look like this.\n\n```python\nfrom sqs_workers import SQSEnv, create_standard_queue\nsqs = SQSEnv()\n\ncreate_standard_queue(sqs, \'emails_dead\')\ncreate_standard_queue(sqs, \'emails\',\n    redrive_policy=sqs.redrive_policy(\'emails_dead\', 3)\n)\n```\n\nIt means "move the message to the email_deadletters queue after four (3 + 1) failed attempts to send it to the recipient."\n\n## Backoff policies\n\nYou can define the backoff policy for the entire environment or specific queue.\n\n```python\nqueue = sqs.queue("emails", backoff_policy=DEFAULT_BACKOFF)\n\n@queue.processor(\'send_email\')\ndef send_email(to, subject, body):\n    print(f"Sending email {subject} to {to}")\n```\n\nThe default policy is the exponential backoff. We recommend setting both the backoff policy and the dead-letter queue to limit the maximum number of execution attempts.\n\nAlternatively, you can set the backoff to IMMEDIATE_RETURN to re-execute the failed task immediately.\n\n```python\nqueue = sqs.queue("emails", backoff_policy=IMMEDIATE_RETURN)\n\n@queue.processor(\'send_email\')\ndef send_email(to, subject, body):\n    print(f"Sending email {subject} to {to}")\n```\n\n## Shutdown policies\n\nWhen launching the queue processor with process_queue(), it\'s possible to optionally set when it has to be stopped.\n\nFollowing shutdown policies are supported:\n\n- IdleShutdown(idle_seconds): return from function when no new tasks are sent for a specific period.\n\n- MaxTasksShutdown(max_tasks): return from function after processing at least max_task jobs. It can be helpful to prevent memory leaks.\n\nThe default policy is NeverShutdown. It\'s also possible to combine two previous policies with OrShutdown or AndShutdown policies or create custom classes for specific behavior.\n\nExample of stopping processing the queue after 5 minutes of inactivity:\n\n```python\nfrom sqs_workers import SQSEnv\nfrom sqs_workers.shutdown_policies import IdleShutdown\n\nsqs = SQSEnv()\nsqs.queue("foo").process_queue(shutdown_policy=IdleShutdown(idle_seconds=300))\n```\n\n## Processing a dead-letter queue by pushing back failed messages\n\nThe most common way to process a dead-letter queue is to fix the main bug causing messages to appear there in the first place and then to re-process these messages again.\n\nWith sqs-workers, in can be done by putting back all the messages from a dead-letter queue back to the main one. While processing the queue, the processor takes every message and pushes it back to the upstream queue with a hard-coded delay of 1 second.\n\nUsage example:\n\n    >>> from sqs_workers import JobQueue\n    >>> from sqs_workers.shutdown_policies IdleShutdown\n    >>> from sqs_workers.deadletter_queue import DeadLetterQueue\n    >>> env = SQSEnv()\n    >>> foo = env.queue("foo")\n    >>> foo_dead = env.queue("foo_dead", DeadLetterQueue.maker(foo))\n    >>> foo_dead.process_queue(shutdown_policy=IdleShutdown(10))\n\nThis code takes all the messages in the foo_dead queue and pushes them back to the foo queue. Then it waits 10 seconds to ensure no new messages appear, and quit.\n\n## Processing a dead-letter queue by executing tasks in-place\n\nInstead of pushing the tasks back to the main queue, you can execute them in place. For this, you need to copy the queue processors from the main queue to the deadletter.\n\nUsage example:\n\n    >>> env = SQSEnv()\n    >>> foo = env.queue("foo")\n    >>> foo_dead = env.queue("foo_dead")\n    >>> foo.copy_processors(foo_dead)\n    >>>\n    >>> from sqs_workers.shutdown_policies IdleShutdown\n    >>> foo_dead.process_queue(shutdown_policy=IdleShutdown(10))\n\nThis code takes all the messages in the foo_dead queue and executes them with processors from the "foo" queue. Then it waits 10 seconds to ensure no new messages appear, and quit.\n\n## Using in unit tests with MemorySession\n\nThere is a special MemorySession, which can be used as a quick and dirty replacement for real queues in unit tests. If you have a function `create_task` which adds some tasks to the queue and you want to test how it works, you can technically write your tests like this:\n\n```python\nfrom sqs_workers import SQSEnv\nenv = SQSEnv()\n\ndef test_task_creation_side_effects():\n    create_task()\n    env.process_batch(\'foo\')\n    ...\n```\n\nThe problem is that your test starts depending on AWS (or localstack) infrastructure, which you don\'t always need. What you can do instead is you can pass MemorySession to your SQSEnv instance.\n\n```python\nfrom sqs_workers import SQSEnv, MemorySession\nenv = SQSEnv(MemorySession())\n```\n\nPlease note that MemorySession has some serious limitations, and may not fit well your use-case. Namely, when you work with MemorySession:\n\n- Redrive policy doesn\'t work\n- There is no differences between standard and FIFO queues\n- FIFO queues don\'t support content-based deduplication\n- Delayed tasks are executed ineffectively: the task is gotten from the queue, and if the time hasn\'t come, the task is put back.\n- API can return slightly different results\n\n## Contributing\n\nPlease see our guide [here](./CONTRIBUTING.md)\n\n## Local Development\n\nWe use Poetry for dependency management & packaging.  Please see [here for setup instructions](https://python-poetry.org/docs/#installation).\n\nOnce you have Poetry installed, you can run the following to install the dependencies in a virtual environment:\n\n```bash\npoetry install\n```\n\n## Testing\n\nWe use pytest to run unittests, and tox to run them for all supported Python versions.\n\nIf you just run `pytest` or `tox`, all tests will be run against AWS, localstack, and MemorySession. You can disable those you don\'t want to use using the pytest `-k` flag, for instance using `-k localstack` or `-k \'not aws\'`.\n\n### Testing with AWS\n\nMake sure you have your boto3 client configured ([ref](https://boto3.readthedocs.io/en/latest/guide/quickstart.html#configuration)) and then run\n\n```bash\npoetry run pytest -k aws\n```\n\nAlternatively, to test all supported versions, run\n\n```bash\npoetry run tox -- -k aws\n```\n\n### Testing with localstack\n\nLocalstack tests should perform faster than testing against AWS, and besides, they work well in offline.\n\nRun [ElasticMQ](https://github.com/softwaremill/elasticmq) and make sure that the SQS endpoint is available by the address localhost:4566:\n\n```bash\ndocker run -p 4566:9324 --rm -it softwaremill/elasticmq-native\n```\n\nThen run\n\n```bash\npoetry run pytest -k localstack\n```\n\nor\n\n```bash\npoetry run tox -- -k localstack\n```\n\n### Testing with MemorySession\n\nMemorySession should be even faster, but has all the limitations documented above. But it can still be useful to test some logic changes.\n\nSimply run\n\n```bash\npoetry run pytest -k memory\n```\n\nor\n\n```bash\npoetry run tox -- -k memory\n```\n\n## Releasing new versions\n\n- Bump version in `pyproject.toml`\n- Update the CHANGELOG\n- Commit the changes with a commit message "Version X.X.X"\n- Tag the current commit with `vX.X.X`\n- Create a new release on GitHub named `vX.X.X`\n- GitHub Actions will publish the new version to PIP for you\n',
    'author': 'Doist Developers',
    'author_email': 'dev@doist.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Doist/sqs-workers',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
import multiprocessing
import warnings
from typing import TYPE_CHECKING, Dict, Optional, Type, Union

import attr
import boto3

from sqs_workers import DEFAULT_BACKOFF, RawQueue, codecs, context, processors
from sqs_workers.batching import BatchingConfiguration, NoBatching
from sqs_workers.core import RedrivePolicy
from sqs_workers.processors import DEFAULT_CONTEXT_VAR
from sqs_workers.queue import GenericQueue, JobQueue
from sqs_workers.shutdown_policies import NeverShutdown

if TYPE_CHECKING:
    from sqs_workers.backoff_policies import BackoffPolicy


logger = logging.getLogger(__name__)


AnyQueue = Union[GenericQueue, JobQueue, RawQueue]


@attr.s
class SQSEnv(object):
    session = attr.ib(default=boto3)
    queue_prefix = attr.ib(default="")

    # queue-specific settings
    backoff_policy = attr.ib(default=DEFAULT_BACKOFF)

    # jobqueue-specific settings
    processor_maker = attr.ib(default=processors.Processor)
    context_maker = attr.ib(default=context.SQSContext)

    # internal attributes
    context = attr.ib(default=None)
    sqs_client = attr.ib(default=None)
    sqs_resource = attr.ib(default=None)
    queues: Dict[str, AnyQueue] = attr.ib(init=False, factory=dict)

    def __attrs_post_init__(self):
        self.context = self.context_maker()
        self.sqs_client = self.session.client("sqs")
        self.sqs_resource = self.session.resource("sqs")

    def queue(
        self,
        queue_name: str,
        queue_maker: Type[AnyQueue] = JobQueue,
        batching_policy: BatchingConfiguration = NoBatching(),
        backoff_policy: Optional["BackoffPolicy"] = None,
    ) -> AnyQueue:
        """
        Get a queue object, initializing it with queue_maker if necessary.
        """
        if queue_name not in self.queues:
            backoff_policy = backoff_policy or self.backoff_policy
            self.queues[queue_name] = queue_maker(
                env=self,
                name=queue_name,
                batching_policy=batching_policy,
                backoff_policy=backoff_policy,
            )
        return self.queues[queue_name]

    def processor(
        self,
        queue_name: str,
        job_name: str,
        pass_context=False,
        context_var=DEFAULT_CONTEXT_VAR,
    ):
        """
        Decorator to attach processor to all jobs "job_name" of the queue "queue_name".
        """
        q: JobQueue = self.queue(queue_name, queue_maker=JobQueue)  # type: ignore
        return q.processor(
            job_name=job_name, pass_context=pass_context, context_var=context_var
        )

    def raw_processor(self, queue_name: str):
        """
        Decorator to attach raw processor to all jobs of the queue "queue_name".
        """
        q: RawQueue = self.queue(queue_name, queue_maker=RawQueue)  # type: ignore
        return q.raw_processor()

    def add_job(
        self,
        queue_name,
        job_name,
        _content_type=codecs.DEFAULT_CONTENT_TYPE,
        _delay_seconds=None,
        _deduplication_id=None,
        _group_id=None,
        **job_kwargs,
    ):
        """
        Add job to the queue.
        """
        warnings.warn(
            "sqs.add_job() is deprecated. Use sqs.queue(...).add_job() instead",
            DeprecationWarning,
        )
        q = self.queue(queue_name, queue_maker=JobQueue)  # type: JobQueue
        return q.add_job(
            job_name,
            _content_type=_content_type,
            _delay_seconds=_delay_seconds,
            _deduplication_id=_deduplication_id,
            _group_id=_group_id,
            **job_kwargs,
        )

    def process_queues(self, queue_names=None, shutdown_policy_maker=NeverShutdown):
        """
        Use multiprocessing to process multiple queues at once. If queue names
        are not set, process all known queues

        shutdown_policy_maker is an optional callable which doesn't accept any
        arguments and create a new shutdown policy for each queue.

        Can looks somewhat like this:

            lambda: IdleShutdown(idle_seconds=10)
        """
        if not queue_names:
            queue_names = self.get_all_known_queues()
        processes = []
        for queue_name in queue_names:
            queue = self.queue(queue_name)
            p = multiprocessing.Process(
                target=queue.process_queue,
                kwargs={"shutdown_policy": shutdown_policy_maker()},
            )
            p.start()
            processes.append(p)
        for p in processes:
            p.join()

    def get_all_known_queues(self):
        resp = self.sqs_client.list_queues(**{"QueueNamePrefix": self.queue_prefix})
        if "QueueUrls" not in resp:
            return []
        urls = resp["QueueUrls"]
        ret = []
        for url in urls:
            sqs_name = url.rsplit("/", 1)[-1]
            queue_prefix_len = len(self.queue_prefix)
            ret.append(sqs_name[queue_prefix_len:])
        return ret

    def get_sqs_queue_name(self, queue_name):
        """
        Take "high-level" (user-visible) queue name and return SQS
        ("low level") name by simply prefixing it. Used to create namespaces
        for different environments (development, staging, production, etc)
        """
        return f"{self.queue_prefix}{queue_name}"

    def redrive_policy(self, dead_letter_queue_name, max_receive_count):
        return RedrivePolicy(self, dead_letter_queue_name, max_receive_count)



import random


class BackoffPolicy(object):
    def get_visibility_timeout(self, message):
        raise NotImplementedError()


class ConstantBackoff(BackoffPolicy):
    """
    Backoff policy which always returns the message back to the queue
    immediately on failure
    """

    def __init__(self, backoff_value=0):
        self.backoff_value = backoff_value

    def get_visibility_timeout(self, message):
        return self.backoff_value


class ExponentialBackoff(BackoffPolicy):
    """
    Backoff policy which keeps the message hidden from the queue
    with an exponential backoff
    """

    def __init__(self, base=2, min_visibility_timeout=0, max_visbility_timeout=30 * 60):
        self.base = base  # in seconds
        self.min_visibility_timeout = min_visibility_timeout
        self.max_visibility_timeout = max_visbility_timeout

    def get_visibility_timeout(self, message):
        prev_receive_count = int(message.attributes["ApproximateReceiveCount"]) - 1
        mu = self.min_visibility_timeout + (self.base ** prev_receive_count)
        sigma = float(mu) / 10
        visibility_timeout = random.normalvariate(mu, sigma)
        visibility_timeout = max(self.min_visibility_timeout, visibility_timeout)
        visibility_timeout = min(self.max_visibility_timeout, visibility_timeout)
        return int(round(visibility_timeout))


DEFAULT_BACKOFF = ExponentialBackoff()
IMMEDIATE_RETURN = ConstantBackoff()
